import sys, os
import random
import string
import subprocess
import json
import threading
import Queue
import atexit
import SocketServer
from wsgiref import simple_server
from wsgiref.util import shift_path_info


this_dir = os.path.dirname(__file__)
root_dir = os.path.abspath(os.path.dirname(this_dir))

ex_js = os.path.join(this_dir, 'ex.js')
rt_rhino_js = os.path.join(this_dir, 'rt-rhino.js')


class JSError(Exception):

    def __init__(self, name, message, file_name, line_number, stack):
        self.name = name
        self.message = message
        self.file_name = file_name
        self.line_number = line_number
        self.stack = stack

    def __str__(self):
        return "%s(%s,\n%s)" % (self.name, self.message, self.stack)


def find_file(fname):
    cand = os.path.abspath(os.path.join(this_dir, fname))
    if cand.startswith(root_dir) and os.path.isfile(cand):
        return cand
    cand = os.path.abspath(os.path.join(root_dir, fname))
    if cand.startswith(root_dir) and os.path.isfile(cand):
        return cand
    return None


class Rhino(object):

    load_fs_based = True

    def __init__(self):
        self.token = "brdg;%s" % ''.join(random.choice(string.ascii_letters)
                                         for i in xrange(12))
        self.p = subprocess.Popen(['js', '-debug',
                                   '-f', ex_js, rt_rhino_js, self.token],
                                  stdin=subprocess.PIPE,
                                  stdout=subprocess.PIPE,
                                  shell=False)
    def send(self, msg):
        self.p.stdin.write("%s\n" % msg)

    def receive(self):
        return self.p.stdout.readline()

    def finish(self):
        self.p.wait()


class TestingWsgiApp(object):

    def __init__(self, in_queue, out_queue):
        self.in_queue = in_queue
        self.out_queue = out_queue
        self.fallback = None

    def _fallback(self, environ, start_response):
        start_response("404 Not Found", [("content-type", "text/plain")])
        return [':/']

    def __call__(self, environ, start_response):
        if environ['PATH_INFO'].startswith('/+test'):
            shift_path_info(environ)
            path_info, method = environ['PATH_INFO'], environ['REQUEST_METHOD']
            if path_info == '/':
                path_info = '/ld/testing.html'
            if method == 'GET' and path_info.startswith('/ld/'):
                fpath = find_file(path_info[4:])
                if fpath is not None:
                    with open(fpath, 'r') as f:
                        data = f.read()
                    if fpath.endswith('html'):
                        content_type = 'text/html'
                    elif fpath.endswith('.js'):
                        content_type = 'text/javascript'
                    else:
                        content_type = 'application/octet-stream'
                    start_response("200 OK", [("content-type", content_type),
                                              ("pragma", "no-cache"),
                                              ("cache-control", "no-cache")])
                    return [data]
            elif method == 'GET' and path_info == '/cmd':
                msg = self.in_queue.get()
                start_response("200 OK", [("content-type", "application/json"),
                                          ("pragma", "no-cache"),
                                          ("cache-control", "no-cache")])
                return [msg]
            elif method == 'POST' and path_info == '/cmd':
                n = int(environ['CONTENT_LENGTH'])
                msg = environ["wsgi.input"].read(n)
                start_response("200 OK", [("content-type", "application/json")])
                self.out_queue.put(msg)
                return [json.dumps({"ok": True})]
            return self._fallback(environ, start_response)

        fallback = self.fallback
        if fallback is None:
            fallback = self._fallback
        return fallback(environ, start_response)


class Browser(object):

    load_fs_based = False
    token = ''

    def __init__(self, test_browser):
        self.in_queue, self.out_queue = Queue.Queue(), Queue.Queue()
        bind_host = host = os.getenv('TEST_HOST', '127.0.0.1')
        if host not in ('localhost', '127.0.0.1'):
            bind_host = '0.0.0.0'
        port = int(os.getenv('TEST_PORT', 0))
        class _RequestHandler(simple_server.WSGIRequestHandler):
            def log_request(*args):
                pass  # suppress
        class _WSGIServer(simple_server.WSGIServer,
                          SocketServer.ThreadingMixIn):
            pass
        self.server = _WSGIServer((bind_host, port), _RequestHandler)
        self.testing_app = TestingWsgiApp(self.in_queue, self.out_queue)
        self.server.set_app(self.testing_app)
        self.server_thread = threading.Thread(target=self.server.serve_forever,
                                              kwargs=dict(poll_interval=0.01))
        self.server_thread.setDaemon(True)
        self.server_thread.start()
        self.base_url = "http://%s:%d" % (host, self.server.server_address[1])
        url = self.base_url + "/+test/"
        if test_browser == 'manual':
            print >>sys.stderr, "point browser at: %s" % url
        else:
            os.system("%s %s &" % (test_browser, url))

    def finish(self):
        self.server.shutdown()
        self.server_thread.join()
        self.server.socket.close()

    def send(self, msg):
        self.in_queue.put(msg)

    def receive(self):
        res = self.out_queue.get()
        return res


class JsBridge(object):

    def __init__(self):
        test_browser = os.getenv('TEST_BROWSER')
        if test_browser is not None:
            self._bridge = Browser(test_browser)
            self.rhino = False
        else:
            self._bridge = Rhino()
            self._stdout = os.getenv('RHINO_STDOUT') is not None
            self.rhino = True
        self._cbs = {}

    def set_fallback_app(self, app):
        assert not self.rhino
        self._bridge.testing_app.fallback = app

    def get_server(self):
        assert not self.rhino
        return self._bridge.server

    def get_base_url(self):
        assert not self.rhino
        return self._bridge.base_url

    def _submit(self, cmd):
        self._bridge.send(json.dumps(cmd))
        while True:
            line = self._bridge.receive()
            if line.startswith(self._bridge.token):
                res = json.loads(line[len(self._bridge.token):])
                if 'cb' in res:
                    try:
                        ret = self._cbs[res['cb']](*res['args'])
                    except KeyboardInterrupt:
                        raise
                    except Exception, e:
                        cb_ret = {"op": "ret", "exc": e.__class__.__name__,
                                  "message": getattr(e, 'message', repr(e))}
                    else:
                        cb_ret = {"op": "ret", "value": ret}
                    self._bridge.send(json.dumps(cb_ret))
                    continue
                break
            else:
                if self._stdout:
                    print line.rstrip()
        if 'error' in res:
            raise JSError(res['error'], res.get('message', ''),
                          res.get('fileName'), res.get('lineNumber', 0),
                          res.get('stack'))
        return res

    def quit(self):
        self._submit({"op": "quit"})
        self._bridge.finish()

    def eval(self, expr):
        res = self._submit({"op": "eval", "expr": expr})
        return res["value"]

    def load(self, fname):
        if self._bridge.load_fs_based:
            fname = find_file(fname)
        self._submit({"op": "load", "fname": fname})

    def global_names(self):
        res = self._submit({"op": "global_names"})
        return set(res)

    def retain(self, name, expr):
        self._submit({"op": "retain", "expr": expr, "name": name})

    def make_cb(self, name, func):
        self._cbs[name] = func
        self._submit({"op": "callback_def", "name": name})

    def forget_cb(self, name):
        self.forget([name])
        del self._cbs[name]

    def forget(self, names):
        self._submit({"op": "forget", "names": names})

    def forget_all(self):
        self._submit({"op": "forget_all"})


_js_bridge = None


def get_js_bridge():
    global _js_bridge
    if _js_bridge is None:
        _js_bridge = JsBridge()
        atexit.register(_js_bridge.quit)
    return _js_bridge
