import os
from unittest2 import TestCase

from js_bridge import get_js_bridge, JSError


class TestJsBridge(TestCase):

    def setUp(self):
        self.js_bridge = get_js_bridge()

    def test_eval(self):
        val = self.js_bridge.eval("3 + 2")
        self.assertEqual(5, val)

    def test_eval_undefined(self):
        # for simplicity convert undefined to null
        val = self.js_bridge.eval("undefined")
        self.assertIsNone(val)

    def test_load(self):
        self.js_bridge.load('u1db.js')
        val = self.js_bridge.eval('new U1DBDocument("id", "REV", {"a": 1})')
        self.assertEqual({'content': {u'a': 1}, 'doc_id': 'id',
                          'has_conflicts': False, 'rev': 'REV'}, val)

    def test_retain(self):
        self.js_bridge.retain("v0", "3+4")
        val = self.js_bridge.eval("v0")
        self.assertEqual(7, val)

    def test_retain_constructed_on_retain(self):
        self.js_bridge.retain("v0", "3+4")
        self.js_bridge.retain("v1", "v0*2")
        val = self.js_bridge.eval("v1")
        self.assertEqual(7*2, val)

    def test_error(self):
        with self.assertRaises(JSError) as caught:
            self.js_bridge.eval("undef")
        self.assertEqual("ReferenceError", caught.exception.name)

        with self.assertRaises(JSError) as caught:
            self.js_bridge.eval("throw new Error()")
        self.assertEqual("Error", caught.exception.name)

    def test_unknown_error(self):
        with self.assertRaises(JSError) as caught:
            self.js_bridge.eval("throw new (function myErr() {})")
        self.assertEqual("unknownError", caught.exception.name)

    def test_forget(self):
        self.js_bridge.retain("v0", "3+4")
        self.js_bridge.retain("v1", "3*4")
        self.js_bridge.forget(["v1"])
        with self.assertRaises(JSError) as caught:
            self.js_bridge.eval("v1")
        self.assertEqual("ReferenceError", caught.exception.name)

    def test_forget_all(self):
        self.js_bridge.retain("v0", "3+4")
        self.js_bridge.retain("v1", "3*4")
        self.js_bridge.forget_all()
        for varname in ("v0", "v1"):
            with self.assertRaises(JSError) as caught:
                self.js_bridge.eval(varname)
            self.assertEqual("ReferenceError", caught.exception.name)

    def test_global_names(self):
        before = self.js_bridge.global_names()
        self.js_bridge.eval("escaped=1")
        after = self.js_bridge.global_names()
        self.assertEqual(set(['escaped'])|before, after)
        self.js_bridge.load('u1db.js')
        after = self.js_bridge.global_names()
        self.assertIn('U1DB', after)
        self.assertEqual(set([u'escaped', u'U1DBSynchronizer',
                              u'U1DBDocument', u'U1DB', u'U1DBVectorClock',
                              u'U1DBHTTPSyncTarget']), after-before)

    def test_ignore_prints(self):
        res = self.js_bridge.eval("print('one'); print('two'); 3")
        self.assertEqual(3, res)

    def test_callback(self):
        calls = []
        def cb(val1, val2):
            calls.append((val1, val2))
            return val1 + val2
        self.js_bridge.make_cb("callback1", cb)
        res = self.js_bridge.eval("callback1(1, 3); callback1(4, 5)")
        self.assertEqual([(1, 3), (4, 5)], calls)
        self.assertEqual(9, res)
        self.js_bridge.forget_cb("callback1")
        self.assertFalse(self.js_bridge._cbs)
        with self.assertRaises(JSError) as caught:
            self.js_bridge.eval("callback1(0, 1)")
        self.assertEqual("ReferenceError", caught.exception.name)

    def test_callback_exception(self):
        class MyError(Exception):
            pass
        self.js_bridge.retain("MyError",
                          '(function MyError() {this.name = "MyError"})')
        def cb():
            raise MyError()
        self.js_bridge.make_cb("callback2", cb)
        with self.assertRaises(JSError) as caught:
            self.js_bridge.eval("callback2()")
        self.assertEqual("MyError", caught.exception.name)
        self.js_bridge.forget(['MyError'])
        with self.assertRaises(JSError) as caught:
            self.js_bridge.eval("callback2()")
        self.assertEqual("Error", caught.exception.name)


    def test_callback_back_n_forth(self):
        def f(a):
            return self.js_bridge.eval("%r * 2" % a)
        def cb(a):
            return f(a) + 5
        self.js_bridge.make_cb("callback3", cb)
        res = self.js_bridge.eval("callback3(7)")
        self.assertEqual(7*2+5, res)
