
function U1DBSynchronizer(source, sync_target) {
    /*
    Collect the state around synchronizing 2 U1DB replicas.

    Synchronization is bi-directional, in that new items in the source are sent
    to the target, and new items in the target are returned to the source.
    However, it still recognizes that one side is initiating the request. Also,
    at the moment, conflicts are only created in the source.
    */

    var target_replica_uid = null;
    var num_inserted = 0;

    this._insert_doc_from_target = function(doc, replica_gen, trans_id) {
        /*
        Try to insert synced document from target.

        Implements TAKE OTHER semantics: any document from the target
        that is in conflict will be taken as the new official value,
        while the current conflicting value will be stored alongside
        as a conflict. In the process indexes will be updated etc.

        :return: None
        */
        var state = source._put_doc_if_newer(doc, true,
                                             target_replica_uid,
                                             replica_gen,
                                             {replica_trans_id: trans_id})[0];
        switch(state) {
            case "inserted": num_inserted += 1; break;
            case "conflicted": num_inserted += 1; break;
            case "converged": /* magical convergence */
            case "superseded": /* something newer; do at next sync */
                break; /* do nothing */
            default:
                throw (new Error("Invalid state: " + state));
        }
    };

    this._record_sync_info_with_the_target = function(start_generation) {
        /*
        Record our new after sync generation with the target if gapless.

        Any documents received from the target will cause the local
        database to increment its generation. We do not want to send
        them back to the target in a future sync. However, there could
        also be concurrent updates from another process doing eg
        'put_doc' while the sync was running. And we do want to
        synchronize those documents.  We can tell if there was a
        concurrent update by comparing our new generation number
        versus the generation we started, and how many documents we
        inserted from the target. If it matches exactly, then we can
        record with the target that they are fully up to date with our
        new generation.
        */
        var cur = source._get_generation_info();
        var cur_gen = cur[0];
        var cur_trans_id = cur[1];
        if (cur_gen == start_generation + num_inserted && num_inserted > 0) {
            sync_target.record_sync_info(source._get_replica_uid(), cur_gen,
                                        cur_trans_id);
        }
    };

    this.sync = function(opts) {
        /* get target identifier, its current generation,
           and its last-seen database generation for this source */
        var autocreate = U1DB.Utils.Opt(opts, 'autocreate', false);
        var ensure_callback = null;
        var target_replica_uid;
        var info;
        try {
            info = sync_target.get_sync_info(source._get_replica_uid());
        } catch(e) {
            if (e instanceof U1DB.Errors.DatabaseDoesNotExist && autocreate) {
                info = [null /*replica_uid*/, 0, '', 0, ''];
                ensure_callback = function(replica_uid) {
                    target_replica_uid = replica_uid;
                }
            } else {
                throw e;
            }
        }
        target_replica_uid = info[0];
        var target_gen = info[1];
        var target_trans_id = info[2]
        var target_my_gen = info[3];
        var target_my_trans_id = info[4];
        /* validate the generation and transaction id
           the target knows about us */
        source.validate_gen_and_trans_id(target_my_gen, target_my_trans_id);
        /* what's changed since that generation and this current gen */
        var gen_trans_id_changes = source.whats_changed(target_my_gen);
        var my_gen = gen_trans_id_changes[0];
        var changes = gen_trans_id_changes[2];
        /* this source last-seen database generation for the target */
        var target_last_known;
        if (target_replica_uid != null) {
            target_last_known = source._get_replica_gen_and_trans_id(
                target_replica_uid);
        } else {
            target_last_known = [0, ''];
        }
        var target_last_known_gen = target_last_known[0];
        var target_last_known_trans_id = target_last_known[1];
        /* check if there is nothing to do */
        if (changes.length == 0 &&  target_last_known_gen == target_gen) {
            if (target_trans_id != target_last_known_trans_id) {
                throw U1DB.Errors.InvalidTransactionId();
            }
            return my_gen;
        }
        var changed_doc_ids = [];
        changes.forEach(function(didandchg) {
            changed_doc_ids.push(didandchg[0])
        });
        /* prepare to send all the changed docs */
        var docs_to_send = source.get_docs(changed_doc_ids, {
            check_for_conflicts: false, include_deleted: true
        });
        var docs_by_generation = [];
        for (var i=0; i<changes.length; i++) {
            docs_by_generation.push([docs_to_send[i], changes[i][1],
                                    changes[i][2]]);
        }

        /* exchange documents and try to insert the returned ones with
           the target, return target synced-up-to gen */
        var new_pt = sync_target.sync_exchange(docs_by_generation,
                                               source._get_replica_uid(),
                                               target_last_known_gen,
                                               target_last_known_trans_id,
                                               this._insert_doc_from_target,
                                               ensure_callback);
        /* record target synced-up-to generation including applying
           what we sent */
        source._set_replica_gen_and_trans_id(target_replica_uid,
                                             new_pt[0], // new_gen
                                             new_pt[1]); // new_trans_id
        /* if gapless, record current reached generation with target */
        this._record_sync_info_with_the_target(my_gen);
        return my_gen;
    };
};

function U1DBDocument(doc_id, rev, obj, has_conflicts) {
    this.doc_id = doc_id;
    this.rev = rev;
    this.content = obj;
    this.has_conflicts = !!has_conflicts;
}

U1DBDocument.prototype.toString = function() {
    var did = this.doc_id;
    if (did.length > 10) {
        did = this.doc_id.substr(0,3) + "..." + this.doc_id.substr(-3);
    }
    var rev = this.rev;
    if (rev&&rev.length > 12) {
        rev = this.rev.substr(0,3) + "..." + this.rev.substr(-5);
    }
    return "[Document(" + did + ", " + rev +
        (this.has_conflicts ? ", conflicted" : "") + ", " +
        JSON.stringify(this.content) + ")]";
};

U1DBDocument.prototype.is_tombstone = function() {
    return this.content == null;
}

U1DBDocument.prototype.same_content_as = function(doc) {
    return U1DB.Utils.deepEqual(this.content, doc.content);
}

U1DBDocument.prototype.get_json = function() {
    if (this.content == null) {
        return null;
    }
    return JSON.stringify(this.content);
}

U1DBDocument.prototype.set_json = function(json) {
    if (json == null) {
        this.content = null;
    } else {
        this.content = JSON.parse(json);
    }
}


function U1DBVectorClock(value) {
    this.values = this._expand(value);
}

U1DBVectorClock.prototype.as_str = function() {
    var values = this.values;
    var s = [];
    for (var k in values) {
        s.push(k + ":" + values[k]);
    }
    s.sort();
    return s.join("|");
}

U1DBVectorClock.prototype.toString = function() {
    return "[VectorClock " + this.as_str() + "]";
};

U1DBVectorClock.prototype._expand = function(value) {
    var result = {};
    if (value === null) {
        return result
    }
    value.split('|').forEach(function(replica_info) {
        var parts = replica_info.split(":");
        var replica_uid = parts[0];
        var counter = parts[1];
        counter = parseInt(counter, 10);
        result[replica_uid] = counter;
    });
    return result
};

U1DBVectorClock.prototype._values = function() {
    var values = this.values;
    var clone = {};
    for (var k in values) clone[k] = values[k]; // shallow copy
    return clone;
}

U1DBVectorClock.prototype._isEmptyDict = function(dict) {
    for (var k in dict) return false;
    return true;
}

U1DBVectorClock.prototype.is_newer = function(other) {
    var values = this.values;
    if (this._isEmptyDict(values)) {
        return false;
    }
    var other_expand = other._values();
    if (this._isEmptyDict(other_expand)) {
        return true;
    }
    var this_is_newer = false, other_value;
    for (var key in values) {
        var value = values[key];
        if (other_expand[key] !== undefined) {
            other_value = other_expand[key];
            delete other_expand[key];
            if (other_value > value) {
                return false;
            } else if (other_value < value) {
                this_is_newer = true;
            }
            // else they are equal, so carry on
        } else {
            this_is_newer = true;
        }
    }
    if (!this._isEmptyDict(other_expand)) return false;
    return this_is_newer;
};

U1DBVectorClock.prototype.increment = function(replica_uid) {
    var values = this.values;
    if (values[replica_uid] === undefined) {
        values[replica_uid] = 1;
    } else {
        values[replica_uid] += 1;
    }
};

U1DBVectorClock.prototype.maximize = function(other_vcr) {
    var values = this.values;
    var counter, this_counter;
    var other_vcr_values = other_vcr._values();
    for (var replica_uid in other_vcr_values) {
        counter = other_vcr_values[replica_uid];
        if (values[replica_uid] === undefined) {
            values[replica_uid] = counter;
        } else {
            this_counter = values[replica_uid];
            if (this_counter < counter) {
                values[replica_uid] = counter;
            }
        }
    };
};


function U1DB(db_name, replica_uid_is_db_name) {

    var that = this;

    /* "import" */
    var deepEqual = U1DB.Utils.deepEqual;
    var Opt = U1DB.Utils.Opt;
    var VarArgs = U1DB.Utils.VarArgs;

    this.toString = function() {
        return "[Database " + db_name + "]";
    };

    /* Internal functions */

    this.debug_prints = false;

    var getItem, setItem, deleteItem, wipe, internalStorage;
    internalStorage = {};
    if (typeof(window) === "undefined") {
        setItem = function(keyname, did, value) {
            internalStorage[["u1db", db_name, keyname, did].join("-")] = value;
        }
        getItem = function(keyname, did) {
            var data = internalStorage[["u1db", db_name, keyname, did].join("-")];
            if (data == null) {
                return null;
            }
            return data;
        }
        deleteItem = function(keyname, did) {
            delete internalStorage[["u1db", db_name, keyname, did].join("-")];
        }
        wipe = function() {
            internalStorage = {};
        }
    } else {
        setItem = function(keyname, did, value) {
            return window.localStorage.setItem(["u1db", db_name, keyname, did].join("-"), value);
        };
        getItem = function(keyname, did) {
            return window.localStorage.getItem(["u1db", db_name, keyname, did].join("-"));
        };
        deleteItem = function(keyname, did) {
            return window.localStorage.removeItem(["u1db", db_name, keyname, did].join("-"));
        };
        wipe = function() {
            var beginning = "u1db-" + db_name + "-";
            for (var i=window.localStorage.length-1; i>=0; i--) { // loop backwards because we alter the loop inside it
                var key = window.localStorage.key(i);
                if (key.substr(0, beginning.length) == beginning) {
                    window.localStorage.removeItem(key);
                }
            }
        }
    }

    /* https://gist.github.com/982883 */
    var UUIDv4 = function b(a){return a?(a^Math.random()*16>>a/4).toString(16):([1e7]+-1e3+-4e3+-8e3+-1e11).replace(/[018]/g,b)};
    var db_replica_uid = getItem("REPLICA", "UID");
    if (db_replica_uid === null) {
        if (replica_uid_is_db_name) {
            db_replica_uid = db_name;
        } else {
            db_replica_uid = UUIDv4();
        }
        setItem("REPLICA", "UID", db_replica_uid);
    }

    var bumpRev = function(doc) {
        var revlist = getItem("rev", doc.doc_id);
        if (revlist != doc.rev && doc.rev !== null) {
            throw (new Error("Unbumped rev (" + revlist + ") not equal to doc rev (" + doc.rev + ")"));
        }
        var vcr = new U1DBVectorClock(revlist);
        vcr.increment(db_replica_uid);
        doc.rev = vcr.as_str();
        setItem("rev", doc.doc_id, doc.rev);
    };

    var getIndex = function(index_name, sorted) {
        var index_obj = JSON.parse(getItem("index", index_name));
        var entries = index_obj.entries;
        if (entries.length != 0 &&  sorted && !index_obj.sorted) {
            setIndex(index_name, entries, true);
        }
        return entries;
    }

    var setIndex = function(index_name, entries, do_sort) {
        var index_obj;
        if (do_sort) {
            entries.sort(compareIndexEntries);
            index_obj = {sorted: true, entries: entries};
        } else {
            index_obj = {sorted: false, entries: entries};
        }
        setItem("index", index_name, JSON.stringify(index_obj));
    }

    var updateIndexes = function(doc) {
        /* Updates all indexes for one doc */
        var indexes = getIndexDefs()
        for (var indexname in indexes) {
            var index_def = indexes[indexname];
            var index_data = getIndex(indexname, false);
            // remove all existing keys for this doc
            var new_index_data = [];
            index_data.forEach(function(keydocid) {
                if (keydocid[1] != doc.doc_id) new_index_data.push(keydocid);
            });
            if (doc.content == null) {
                // skip calculating new keys for it
                if (that.debug_prints) {
                    print("Removing doc " + doc.doc_id + " from index");
                }
            } else {
                var index_keys = getIndexEntriesForDoc(index_def, doc);
                if (that.debug_prints) print("uI eval: " + JSON.stringify(index_keys));
                if (index_keys !== null) {
                    index_keys.forEach(function(index_key) {
                        if (index_key !== null) {
                            new_index_data.push([index_key, doc.doc_id]);
                        }
                    });
                }
            }
            if (that.debug_prints) print("Setting index " + indexname + " to " + JSON.stringify(new_index_data));
            setIndex(indexname, new_index_data, false);
        };
    };

    var updateIndexForAllDocs = function(indexname, index_def) {
        /* Updates one index for all docs */
        if (that.debug_prints) print("uIFAD: " + JSON.stringify([indexname, index_def]));
        var result = [];
        forEachLiveDoc(function(doc) {
            var index_keys = getIndexEntriesForDoc(index_def, doc);
            if (that.debug_prints) print("uIFAD eval: " + JSON.stringify(index_keys));
            if (index_keys !== null) {
                index_keys.forEach(function(index_key) {
                    if (index_key !== null) {
                        result.push([index_key, doc.doc_id]);
                    }
                });
            }
        }, false);
        setIndex(indexname, result, false);
    };

    var P_NONEXISTENT = 1, P_LIVE = 2, P_DEAD = 3;

    var _getDocList = function() {
        var doclist = getItem("doclist", "");
        if (doclist === null) {
            return [];
        } else {
            return JSON.parse(doclist);
        }
    }

    var trackDoc = function(doc, previously) {
        if (previously == P_NONEXISTENT) {
            var doclist = _getDocList()
            doclist.push(doc.doc_id);
            setItem("doclist", "", JSON.stringify(doclist));
        }
    };

    var forEachDoc = function(on_doc, check_for_conflicts) {
        var doclist = _getDocList();
        doclist.forEach(function(doc_id) {
            on_doc(getDoc(doc_id, check_for_conflicts));
        })
    };

    var forEachLiveDoc = function(on_doc, check_for_conflicts) {
        var doclist = _getDocList();
        doclist.forEach(function(doc_id) {
            var doc = getDoc(doc_id, check_for_conflicts);
            if (doc.content == null) {
                return;
            }
            on_doc(doc);
        })
    };

    function getter(term) {
        if (typeof term == "object") {
            return term;
        }
        if (term.slice(-1) == '.') {
            throw new U1DB.Errors.IndexDefinitionParseError();
        }
        return {'op': 'get', 'args': term.split('.')}
    }

    this.getter = getter;

    this.Transformations = {
        number: {arity: 2, args: [getter, parseInt]},
        split_words: {arity: 1, args: [getter]},
        lower: {arity: 1, args: [getter]},
        upper: {arity: 1, args: [getter]},
        bool: {arity: 1, args: [getter]},
        combine: {arity: -1, args: [getter]}
    };

    this.TransformationFunctions = {
        combine: function(/* *vals */) {
            var out = [];
            for(var i=0; i < arguments.length; i++) {
                out = out.concat(arguments[i]);
            };
            return out;
        },
        lower: function(vals) {
            var out = [];
            for (var i=0; i<vals.length; i++) {
                var s = vals[i];
                if (typeof s != "string") {
                    continue;
                }
                out.push(s.toLowerCase());
            };
            return out;
        },
        upper: function(vals) {
            var out = [];
            for (var i=0; i<vals.length; i++) {
                var s = vals[i];
                if (typeof s != "string") {
                    continue;
                }
                out.push(s.toUpperCase());
            };
            return out;
        },
        split_words: function(vals) {
            var out = [];
            for (var i=0; i<vals.length; i++) {
                var s = vals[i];
                if (typeof s != "string") {
                    continue;
                }
                out = out.concat(s.split(/\s+/).filter(
                    function(bit) {return bit != ""}));
            }
            return out;
        },
        bool: function(vals) {
            var out = [];
            for (var i=0; i<vals.length; i++) {
                var bool = vals[i];
                if (typeof bool != "boolean") {
                    continue;
                }
                var val = "0";
                if (bool === true) val = "1";
                out.push(val);
            };
            return out;
        },
        number: function(vals, size) {
            var out = [];
            var sizeofzeroes = (new Array(size+1)).join("0");
            for (var i=0; i<vals.length; i++) {
                var num = vals[i];
                if (typeof num != "number") {
                    continue;
                }
                var num_as_string = "" + num;
                var paddednum = sizeofzeroes + num;
                if (num_as_string.length > size) {
                    out.push(num);
                } else {
                    out.push(paddednum.substr(paddednum.length - size));
                }
            };
            return out;
        }
    };

    function strip(s) {
        return s.replace(/^\s+/, "").replace(/\s+$/, "");
    }

    var parseIndexExpression = function(expression) {
        var tokens = [];
        var token;
        while(expression) {
            var delim = expression.search(/\(|\)|,/);
            if (delim == 0) {
                tokens.push(expression.charAt(0))
                expression = expression.slice(1);
                continue;
            } else if (delim == -1) {
                token = expression;
                expression = null;
            } else {
                token = expression.slice(0, delim);
                expression = expression.slice(delim)
            }
            token = strip(token);
            if (token != "") {
                tokens.push(token)
            }
        }

        function parse_term() {
            var term = tokens.shift();
            if (term == null || term == ',' || term == '(' || term == ')') {
                throw new U1DB.Errors.IndexDefinitionParseError();
            }
            if (tokens[0] == '(') {
                return parse_op(term)
            }
            return term
        }

        function  parse_op(op_name) {
            tokens.shift(); // '('
            var op = that.Transformations[op_name];
            if (op == null) {
                throw new U1DB.Errors.IndexDefinitionParseError();
            }
            var args = [];
            while (true) {
                args.push(parse_term());
                var sep = tokens.shift();
                if (sep == ')') {
                    break;
                } else if (sep != ',') {
                    throw new U1DB.Errors.IndexDefinitionParseError();
                }
            }

            var parsed = [];
            for (var i = 0; i < args.length; i++) {
                var inner;
                try {
                    inner = op.args[Math.min(i, op.args.length-1)](args[i]);
                } catch(e) {
                    throw new U1DB.Errors.IndexDefinitionParseError();
                }
                parsed.push(inner);
            }
            if (op.arity != -1 && parsed.length != op.arity) {
                throw new U1DB.Errors.IndexDefinitionParseError();
            }
            return {op: op_name, args: parsed};
        }

        var term = getter(parse_term())
        if(tokens[0] != null) {
            throw new U1DB.Errors.IndexDefinitionParseError();
        }

        return term;
    };

    function parseIndexExpressions(index_expressions) {
        var trees = [];
        index_expressions.forEach(function (index_expression) {
            trees.push(parseIndexExpression(index_expression));
        })
        return trees;
    }

    this._parseIndexExpressions = parseIndexExpressions;

    function doGet(content, subfields, index) {
        index = index || 0;
        if (typeof content != "object") {
            return [];
        }
        var val = content[subfields[index]];
        if (val == null) {
            return [];
        }
        if (index < subfields.length-1) {
            if (val instanceof Array) {
                var results = [];
                val.forEach(function (item) {
                    results = results.concat(doGet(item, subfields, index+1));
                })
                return results;
            }
            if (typeof val == "object") {
                return doGet(val, subfields, index+1);
            }
            return [];
        }
        if (val instanceof Array) {
            /* Strip anything in the list that isn't a simple type */
            return val.filter(function(v) {return typeof v != "object"})
        }
        if (typeof val == "object") {
            return [];
        }
        return [val];
    }

    function evalTree(tree, content) {
        if (tree.op == 'get') {
            return doGet(content, tree.args);
        }
        var vals = [];
        tree.args.forEach(function(arg) {
            if (typeof arg == 'object' && 'op' in arg) {
                arg = evalTree(arg, content);
            }
            vals.push(arg);
        })
        return that.TransformationFunctions[tree.op].apply(null, vals);
    }

    function getIndexEntriesForDoc(index_def, doc) {
        var trees = index_def.trees;
        var all_rows = [[]];
        for (var j=0; j<trees.length; j++) {
            var keys = evalTree(trees[j], doc.content);
            if (keys == null) {
                return [];
            }
            var new_rows = [];
            keys.forEach(function(key) {
                all_rows.forEach(function(row) {
                    new_rows.push(row.concat([key]))
                })
            })
            all_rows = new_rows;
        }
        return all_rows;
    }

    this._getIndexEntriesForDoc = getIndexEntriesForDoc;

    var getTransactionLog = function() {
        var tl = getItem("transactionlog", "transactionlog");
        if (tl === null) {
            tl = [];
        } else {
            tl = JSON.parse(tl);
        };
        return tl;
    };

    var TRANS_ID = '';

    // only for use by test suite
    this._get_transaction_log = function() {
        return getTransactionLog().map(function(did) {return [did, TRANS_ID]});
    }

    var get_generation = function() {
        var tl = getTransactionLog();
        return tl.length;
    };

    this._get_generation = function() { return get_generation(); }

    this._get_generation_info = function() {
        return [get_generation(), ''];
    }

    var updateTransactionLog = function(doc) {
        var tl = getTransactionLog();
        tl.push(doc.doc_id);
        setItem("transactionlog", "transactionlog", JSON.stringify(tl));
    };

    this._debug_dump = function() {
        /* Debugging function: dump the whole database */
        print("== Documents ==");
        forEachDoc(print, true);
        print("== Indexes ==");
        var indexlist = getIndexDefs()
        for (var indexname in indexlist) {
            print("  === " + indexname + ": " + JSON.stringify(indexlist[indexname]) + " ===");
            var index = getIndex(indexname, false);
            for (var j=0; j<index.length; j++) {
                print("  " + JSON.stringify(index[j][0]) + ":" + JSON.stringify(index[j][1]));
            }
        }
        print("=============");
    };

    this._get_replica_uid = function() { return db_replica_uid; }

    var NIL = "";

    var getDoc = function(doc_id, check_for_conflicts) {
        var data = getItem("data", doc_id);
        if (data == null) return null;
        if (data == NIL) {
            data = null;
        }
        var doc = new U1DBDocument(doc_id, getItem("rev", doc_id));
        doc.set_json(data);
        var conflicts = null;
        if (check_for_conflicts) {
            conflicts = getItem("conflicts", doc_id);
            if (that.debug_prints) print("get_doc: got doc " + doc_id + " which has conflicts " + JSON.stringify(conflicts));
            if (conflicts !== null) doc.has_conflicts = true;
        }
        if (that.debug_prints) print("get_doc: returning " + doc);
        return doc;
    }

    var putDocAndUpdateIndexes = function(doc, bump_rev, previously) {
        if (that.debug_prints) print("Updating doc " + doc.doc_id + " content to " + JSON.stringify(doc.content));
        var json_cont = doc.get_json();
        if (json_cont == null) {
            json_cont = NIL;
        }
        setItem("data", doc.doc_id, json_cont);
        if (bump_rev) {
            bumpRev(doc);
        } else {
            setItem("rev", doc.doc_id, doc.rev);
        }
        updateIndexes(doc);
        trackDoc(doc, previously);
        updateTransactionLog(doc);
    }

    var conflictsAsList = function(doc_id) {
        var conflicts = getItem("conflicts", doc_id);
        if (conflicts === null) {
            return [];
        }
        return JSON.parse(conflicts);
    }

    var replaceConflicts = function(doc, conflicts) {
        if (that.debug_prints) print("Setting conflicts on " + doc.doc_id + " to " + JSON.stringify(conflicts));

        if (conflicts.length == 0) {
            deleteItem("conflicts", doc.doc_id);
            doc.has_conflicts = false;
        } else {
            setItem("conflicts", doc.doc_id, JSON.stringify(conflicts));
            doc.has_conflicts = true;
        }
    }

    var adjustConflicts = function(doc, new_conflict_doc) {
        var conflicts = conflictsAsList(doc.doc_id);
        if (new_conflict_doc) {
            conflicts.push([new_conflict_doc.rev, new_conflict_doc.content]);
        }
        /* prune conflicts */
        var doc_vcr = new U1DBVectorClock(doc.rev);
        var pruned_conflicts = [];
        var autoresolved = false;
        conflicts.forEach(function(conflict) {
            var conf_vcr = new U1DBVectorClock(conflict[0]);
            if (doc_vcr.is_newer(conf_vcr)) {
                return;
            } else if (deepEqual(doc.content, conflict[1])) {
                doc_vcr.maximize(new U1DBVectorClock(conflict[0]));
                autoresolved = true;
            } else {
                pruned_conflicts.push(conflict);
            }
        });
        if (autoresolved) {
            doc_vcr.increment(db_replica_uid);
            doc.rev = doc_vcr.as_str();
        }
        replaceConflicts(doc, pruned_conflicts)
    }

    var DOC_ID_CONTRAINTS = /^[a-zA-Z0-9.%_-]+$/;

    function compareIndexEntries(e1, e2) {
        var k1=e1[0], k2=e2[0];
        /* assumes k2 is longer than or same length as k1 */
        for (var i = 0; i < k1.length; i++) {
            if (k1[i] == k2[i]) {
                continue
            }
            if (k1[i] < k2[i]) {
                return -1;
            }
            return 1;
        }
        if (k1.length < k2.length) {
            return -1;
        }
        return 0;
    }

    this._compareIndexEntries = compareIndexEntries;

    function binarySearch(data, e) {
        var start = 0;
        var end = data.length-1;
        while (end > start) {
            var i = Math.floor((start+end)/2);
            var cmp = compareIndexEntries(e, data[i]);
            if (cmp > 0) {
                start = i+1;
            } else {
                end = i;
            }
        }
        if (compareIndexEntries(e, data[start]) <= 0) {
            return start;
        }
        return start+1;
    }

    this._binarySearch = binarySearch;

    function getIndexDefs() {
        var indexes = getItem("indexes", "indexes");
        if (!indexes) {
            indexes = {};
        } else {
            indexes = JSON.parse(indexes);
        }
        return indexes;
    }

    function getIndexExpressions(index_name) {
        var indexes = getIndexDefs();
        var index_info = indexes[index_name];
        if (index_info == null) {
            throw (new U1DB.Errors.IndexDoesNotExist());
        }
        return index_info.exprs;
    }

    function updateIndexDefs(indexes) {
        setItem("indexes", "indexes", JSON.stringify(indexes));
    }

    /* Public API */

    this.delete_db = function() { wipe(); }

    this.toString = function() { return "[U1DB '" + db_name + "']"; }

    this.close = function() { };
    this.create_doc_from_json = function(json, doc_id_opts) {
        return this.create_doc(JSON.parse(json), doc_id_opts)
    }
    this.create_doc = function(obj, doc_id_opts) {
        var doc_id;
        if (typeof doc_id_opts == "string") {
            doc_id = doc_id_opts;
        } else {
            doc_id = Opt(doc_id_opts, 'doc_id', null);
        }
        if (!doc_id) {
            doc_id = UUIDv4();
        } else if (!DOC_ID_CONTRAINTS.test(doc_id)) {
            throw new U1DB.Errors.InvalidDocId();
        }
        var doc = new U1DBDocument(doc_id, null, obj);
        this.put_doc(doc);
        return doc;
    }
    this.get_doc = function(doc_id, opts) {
        var include_deleted = Opt(opts, 'include_deleted', false);
        var doc = getDoc(doc_id, true);
        if (doc!=null&&doc.content==null&&!include_deleted) {
            return null;
        }
        return doc
    }
    this.delete_doc = function(doc) {
        var currev = getItem("rev", doc.doc_id);
        if (currev === null) {
            throw (new U1DB.Errors.DocumentDoesNotExist());
        }
        if (currev != doc.rev) {
            throw (new U1DB.Errors.RevisionConflict());
        }
        if (getItem("data", doc.doc_id) == NIL) {
            throw (new U1DB.Errors.DocumentAlreadyDeleted());
        }
        if (getItem("conflicts", doc.doc_id) !== null) {
            throw (new U1DB.Errors.ConflictedDoc());
        }
        doc.content = null;
        putDocAndUpdateIndexes(doc, true, P_LIVE);
    }
    this.put_doc = function(doc) {
        if (!DOC_ID_CONTRAINTS.test(doc.doc_id)) {
            throw (new U1DB.Errors.InvalidDocId());
        }
        if (doc.doc_id === null) {
            throw (new U1DB.Errors.InvalidDocId());
        }
        var previously = P_LIVE;
        var currev = getItem("rev", doc.doc_id);
        if (currev != null) {
            if (getItem("data", doc.doc_id) == NIL) {
                previously = P_DEAD;
            }
        } else {
            previously = P_NONEXISTENT;
        }
        if (currev != doc.rev) {
            if (previously != P_DEAD || doc.rev != null) {
                throw (new U1DB.Errors.RevisionConflict());
            }
        }
        if (getItem("conflicts", doc.doc_id) !== null) {
            throw (new U1DB.Errors.ConflictedDoc());
        }
        var bump_rev;
        if (currev === null) {
            if (doc.rev == null) doc.rev = db_replica_uid + ":1";
            bump_rev = false;
        } else {
            bump_rev = true;
        }
        if (that.debug_prints) print("Updating doc " + doc.doc_id + " content to " + JSON.stringify(doc.content));
        putDocAndUpdateIndexes(doc, bump_rev, previously);
        return doc.rev;
    };

    this.get_docs = function(list, opts) {
        var check_for_conflicts = Opt(opts, 'check_for_conflicts', true);
        var include_deleted = Opt(opts, 'include_deleted', false);
        var ret = [];
        for (var i=0; i<list.length; i++) {
            var doc = getDoc(list[i], check_for_conflicts);
            if (doc!=null&&(doc.content!=null||include_deleted)) {
                ret.push(doc);
            }
        }
        return ret;
    };

    this.get_all_docs = function(opts) {
        var include_deleted = Opt(opts, 'include_deleted', false);
        var for_all = include_deleted?forEachDoc:forEachLiveDoc;
        var ret = [];
        var gen = this._get_generation();
        for_all(function (doc) {
            ret.push(doc);
        }, true);
        return [gen,ret];
    };

    this.validate_gen_and_trans_id = function(generation, trans_id) {
        if (trans_id != TRANS_ID) {
            throw new U1DB.Errors.InvalidTransactionId();
        }
    }

    this._validate_source = function(other_replica_uid, other_generation,
                                     other_transaction_id) {
        var old_gen_trans_id = this._get_replica_gen_and_trans_id(
            other_replica_uid);
        var old_generation = old_gen_trans_id[0];
        var old_transaction_id = old_gen_trans_id[1];
        if (other_generation < old_generation) {
            throw new U1DB.Errors.InvalidGeneration();
        }
        if (other_generation > old_generation) {
            return;
        }
        if (other_transaction_id == old_transaction_id) {
            return;
        }
        throw new U1DB.Errors.InvalidTransactionId();
    }

    this._put_doc_if_newer = function(doc, save_conflict,
                                      replica_uid, replica_gen,
                                      opts) {
        var replica_trans_id = Opt(opts, 'replica_trans_id', '');
        this._validate_source(replica_uid, replica_gen, replica_trans_id);
        var cur_doc = getDoc(doc.doc_id);
        var doc_vcr = new U1DBVectorClock(doc.rev);
        var cur_vcr, cur_content, state, previously;
        if (cur_doc === null) {
            cur_vcr = new U1DBVectorClock(null);
            cur_content = null;
            previously = P_NONEXISTENT;
        } else {
            cur_vcr = new U1DBVectorClock(cur_doc.rev);
            cur_content = cur_doc.content;
            previously = cur_content==null?P_DEAD:P_LIVE;
        }
        if (this.debug_prints) print(cur_vcr, doc_vcr);
        cur_vcr.is_newer(doc_vcr);
        if (this.debug_prints) print(cur_vcr, doc_vcr);
        if (doc_vcr.is_newer(cur_vcr)) {
            var rev = doc.rev;
            adjustConflicts(doc);
            putDocAndUpdateIndexes(doc, false, previously);
            if (doc.rev != rev) {
                // conflicts have been autoresolved
                state = "superseded";
            } else {
                state = "inserted";
            }
        } else if (doc.rev == cur_doc.rev) {
            // magical convergence
            state = "converged";
        } else if (cur_vcr.is_newer(doc_vcr)) {
            state = "superseded";
        } else if (deepEqual(doc.content, cur_content)) {
            doc_vcr.maximize(cur_vcr)
            doc_vcr.increment(db_replica_uid);
            doc.rev = doc_vcr.as_str();
            putDocAndUpdateIndexes(doc, false, previously);
            state = 'superseded'
        } else {
            state = "conflicted";
            if (save_conflict) {
                adjustConflicts(doc, cur_doc);
                putDocAndUpdateIndexes(doc, false, previously);
            }
        }
        if (replica_uid != null && replica_gen != null) {
            this._set_replica_gen_and_trans_id(replica_uid, replica_gen,
                                               replica_trans_id);
        }
        return [state, get_generation()];
    };

    this.get_doc_conflicts = function(doc_id) {
        var conflicts = conflictsAsList(doc_id);
        if (conflicts.length == 0) {
            return [];
        }
        var result = [getDoc(doc_id)];
        result[0].has_conflicts = true;
        conflicts.forEach(function(conflict) {
            var doc_rev = conflict[0];
            var doc_content = conflict[1];
            var thisdoc = new U1DBDocument(doc_id, doc_rev, doc_content);
            result.push(thisdoc);
        });
        return result;
    };

    this.resolve_doc = function(doc, conflicted_doc_revs) {
        var cur_doc = getDoc(doc.doc_id);
        var cur_rev, previously;
        if (cur_doc == null) {
            cur_rev = null;
            previously = P_NONEXISTENT;
        } else {
            cur_rev = cur_doc.rev;
            previously = cur_doc.content==null?P_DEAD:P_LIVE;
        }
        // ensure_maximal_rev
        var vcr = new U1DBVectorClock(cur_rev);
        conflicted_doc_revs.forEach(function(cdr) {
            vcr.maximize(new U1DBVectorClock(cdr));
        });
        vcr.increment(db_replica_uid);
        var new_rev = vcr.as_str();

        var superseded_revs = {};
        conflicted_doc_revs.forEach(function(cdr) {
            superseded_revs[cdr] = "done";
        });
        var remaining_conflicts = [];
        var cur_conflicts = conflictsAsList(doc.doc_id);
        cur_conflicts.forEach(function(rev_and_content) {
            if (rev_and_content[0] in superseded_revs) {
                return;
            } else {
                remaining_conflicts.push(rev_and_content);
            }
        });
        doc.rev = new_rev;
        if (cur_rev in superseded_revs) {
            putDocAndUpdateIndexes(doc, false, previously);
        } else {
            remaining_conflicts.push([new_rev, doc.content]);
        }
        replaceConflicts(doc, remaining_conflicts);
    };

    this.whats_changed = function(old_generation) {
        if(typeof(old_generation) != "number") {
            old_generation = Opt(old_generation, 'old_generation', 0);
        }
        var changes = [];
        var tl = getTransactionLog();
        var relevant_tail = tl.slice(old_generation);
        var cur_generation = old_generation + relevant_tail.length;
        var seen = {};
        var generation = cur_generation;
        relevant_tail.reverse();
        relevant_tail.forEach(function(doc_id) {
            if (seen[doc_id] === undefined) {
                changes.push([doc_id, generation, TRANS_ID]);
                seen[doc_id] = "done";
            }
            generation -= 1;
        });
        changes.reverse();
        return [cur_generation, TRANS_ID, changes];
    };

    this._set_replica_gen_and_trans_id = function(other_replica_uid,
                                                  other_generation,
                                                  other_transaction_id) {
        var bits = JSON.stringify([other_generation, other_transaction_id])
        setItem("sync_generation", other_replica_uid, bits);
    };

    this._get_replica_gen_and_trans_id = function(other_replica_uid) {
        var bits = getItem("sync_generation", other_replica_uid);
        if (bits == null) {
            return [0, ''];
        }
        return JSON.parse(bits);
    };

    this.get_sync_target = function() {
        return null;
// xxx following code would need updating/fixing!
//        var db = that;
//        var trace_hook = null;
//        var SyncTarget = function() {
//            /* SyncExchange */
//            var changes_to_return, seen_ids = {};
//            var insert_doc_from_source = function(doc, source_gen, source_replica_uid, main_source_gen) {
//                /*  Try to insert synced document from source.
//                    Conflicting documents are not inserted but will be sent over
//                    to the sync source.
//                    It keeps track of progress by storing the document source
//                    generation as well.
//                    The 1st step of a sync exchange is to call this repeatedly to
//                    try insert all incoming documents from the source. */
//                var state_at_gen = db._put_doc_if_newer(doc, false,
//                                                        source_replica_uid,
//                                                        source_gen);
//                if (db.debug_prints) print("IDFS: state for doc", doc.doc_id, " is ", state_at_gen);
//                switch(state_at_gen[0]) {
//                    case "inserted":
//                        seen_ids[doc.doc_id] = state_at_gen[1];
//                        break;
//                    case "converged":
//                        seen_ids[doc.doc_id] = state_at_gen[1];
//                        break;
//                    case "superseded":
//                        break;
//                    case "conflicted":
//                        break;
//                    default:
//                        throw (new Error("Bad state from p_d_i_n"));
//                }
//            };
//            var find_changes_to_return = function(source_last_known_generation) {
//                trace_hook && trace_hook("before whats_changed");
//                var wc = db.whats_changed(source_last_known_generation);
//                trace_hook && trace_hook("after whats_changed");
//                var new_gen = wc[0];
//                changes_to_return = [];
//                wc[1].forEach(function(change) {
//                    var doc_id = change[0];
//                    var gen = change[1];
//                    var seen = seen_ids[doc_id];
//                    if (seen === undefined || seen < gen) {
//                        changes_to_return.push([doc_id, gen]);
//                    }
//                });
//                return new_gen;
//            };
//            var return_docs = function(return_doc_cb) {
//                var changed_doc_ids = [];
//                changes_to_return.forEach(function(change) {
//                    changed_doc_ids.push(change[0]);
//                });
//                trace_hook && trace_hook("before get_docs");
//                var docs = db.get_docs(changed_doc_ids, {
//                    check_for_conflicts: false, include_deleted: true
//                });
//                if (that.debug_prints) print("CHANGES TO RETURN" + JSON.stringify(changes_to_return));
//                if (that.debug_prints) print("CHANGED_DOC_IDS" + JSON.stringify(changed_doc_ids));
//                if (that.debug_prints) print("DOCS TO RETURN" + JSON.stringify(docs));
//                for (var i=0; i<changes_to_return.length; i++) {
//                    return_doc_cb(docs[i], changes_to_return[i][1]);
//                }
//            };
//
//            /* SyncTarget */
//            this.get_sync_info = function(source_replica_uid) {
//                var source_gen = db.get_sync_generation(source_replica_uid);
//                return [replica_uid, getTransactionLog().length, source_gen];
//            };
//            this.record_sync_info = function(source_replica_uid, source_replica_generation) {
//                db.set_sync_generation(source_replica_uid, source_replica_generation);
//            };
//            this.sync_exchange = function(docs_by_generations, source_replica_uid,
//                    last_known_generation, return_doc_cb) {
//                // 1st step: try to insert incoming docs and record progress
//                if (db.debug_prints) print("in sync_exchange with docs_by_generations" + JSON.stringify(docs_by_generations));
//                docs_by_generations.forEach(function(doc_and_doc_gen) {
//                    insert_doc_from_source(doc_and_doc_gen[0], doc_and_doc_gen[1],
//                        source_replica_uid, last_known_generation);
//                });
//                // 2nd step: find changed docs (including conflicts) to return
//                var new_gen = find_changes_to_return(last_known_generation);
//                // final step: return docs and record source replica sync point
//                return_docs(return_doc_cb);
//                return new_gen;
//            }
//            this._set_trace_hook = function(hook) { trace_hook = hook; }
//        }
//        return new SyncTarget();
    };

    this.create_index = function(name) {
        var index_expressions = VarArgs(1, arguments);
        var indexes = getIndexDefs();
        if (name in indexes) {
            var cur_expressions = getIndexExpressions(name);
            if (!deepEqual(cur_expressions, index_expressions)) {
                throw new U1DB.Errors.IndexNameTakenError();
            }
        }
        var index_def = {exprs: index_expressions,
                         trees: parseIndexExpressions(index_expressions)};
        indexes[name] = index_def;
        updateIndexDefs(indexes)
        updateIndexForAllDocs(name, index_def);
    };

    this.delete_index = function(name) {
        if (that.debug_prints) print("delete_index: " + name);
        var indexes = getIndexDefs()
        if (indexes[name] == null) {
            // idempotent
            return
        }
        delete indexes[name];
        updateIndexDefs(indexes)
        deleteItem("index", name);
    };

    this.list_indexes = function() {
        var indexes = getIndexDefs();
        var ret = [];
        for (var key in indexes) {
            ret.push([key, indexes[key].exprs]);
        }
        return ret;
    };

    this.get_index_keys = function(name) {
        getIndexExpressions(name); // check for existence
        /* An index is stored as [ [key, doc_id], ... ] */
        var ret = [];
        var seen = {}
        var index = getIndex(name, false);
        index.forEach(function(actual_index_key_and_doc_id) {
            var key = actual_index_key_and_doc_id[0];
            if (key in seen) {
                return;
            }
            ret.push(key);
            seen[key] = null;
        });
        return ret;
    };

    /* index querying */

    function deGlob(index_keys, index_exprs) {
        /* First, check we've been passed the correct number of keys */
        if (index_exprs.length != index_keys.length) {
            throw (new U1DB.Errors.InvalidValueForIndex());
        }

        var prefix = []
        var glob = false;
        var wild = false;
        for (var j=0; j < index_keys.length; j++) {
            if(index_keys[j].match(/(.*)\*$/) != null) {
                if (glob) {
                    if (index_keys[j] != "*") {
                        throw new U1DB.Errors.InvalidGlobbing();
                    }
                } else {
                    glob = true;
                    var head = index_keys[j].slice(0, -1)
                    if (head != "") {
                        prefix.push(head);
                        wild = true;
                    }
                }
            } else {
                if (glob) {
                    throw new U1DB.Errors.InvalidGlobbing();
                }
                prefix.push(index_keys[j]);
            }
        }

        return [prefix, wild];
    }

    function exactPrefix(prefix) {
        /* used to generate the upper bound for exact full and
           prefix searches */
        var exact_prefix = prefix.slice(0);
        var last = exact_prefix.length-1;
        exact_prefix[last] = exact_prefix[last] + String.fromCharCode(0);
        return exact_prefix
    }

    this._exactPrefix = exactPrefix;

    function starPrefix(prefix) {
        /* used to generate the upper bound for wildcard prefix searches */
        var star_prefix = prefix.slice(0);
        var last = star_prefix.length-1;
        var last_value = star_prefix[last];
        var last_value_last = last_value.length-1;
        var next_char;
        // XXX char(65535) needs special casing!
        next_char = String.fromCharCode(last_value.charCodeAt(
            last_value_last)+1);
        star_prefix[last] = last_value.slice(0, -1) + next_char;
        return star_prefix;
    }

    this._starPrefix = starPrefix;

    function getRangeFromIndex(index_name, start_prefix_wild, end_prefix_wild) {
        var index = getIndex(index_name, true);
        if (index.length == 0) {
            return [];
        }

        var start_prefix = start_prefix_wild[0];
        var end_prefix = end_prefix_wild[0];

        var first_umatched_value = end_prefix_wild[1]?starPrefix:exactPrefix;

        /* find using binary search the indexes start and end
           of first key >= start_prefix (start)
           and first key >= first_umatched_value(end_prefix) (end)
           return the doc ids in between exclusing end
         */

        var start, end;
        if (start_prefix.length == 0) {
            start = 0;
        } else {
            start = binarySearch(index, [start_prefix, null]);
        }

        if (end_prefix.length == 0) {
            end = index.length;
        } else {
            end = binarySearch(index, [first_umatched_value(end_prefix), null]);
        }

        var ret = [];
        var seen = []

        for(var i = start; i < end; i++) {
            var index_key_and_docid = index[i];
            var index_key = index_key_and_docid[0];
            var doc_id = index_key_and_docid[1];
            if (seen[doc_id] === undefined) {
                ret.push(getDoc(doc_id, true));
                seen[doc_id] = null;
            }
        }

        return ret;
    }

    this.get_from_index = function(index_name) {
        /* index_name is the name of the index to query.
           index_keys is a key tuple to lookup in the index
        */
        var index_keys = VarArgs(1, arguments);
        var index_exprs = getIndexExpressions(index_name);
        var prefix_wild = deGlob(index_keys, index_exprs);

        return getRangeFromIndex(index_name, prefix_wild, prefix_wild);
    };

    function prepareRangePrefix(index_keys, index_exprs) {
        if (index_keys === null) {
            return [[], false];
        } else {
            if (typeof index_keys == "string") {
                index_keys = [index_keys];
            }
            return deGlob(index_keys, index_exprs);
        }
    }

    this.get_range_from_index = function(index_name, start_keys, end_keys) {
        var index_exprs = getIndexExpressions(index_name);

        var start_prefix_wild, end_prefix_wild;
        start_prefix_wild = prepareRangePrefix(start_keys, index_exprs);
        end_prefix_wild = prepareRangePrefix(end_keys, index_exprs);

        return getRangeFromIndex(index_name, start_prefix_wild, end_prefix_wild)
    };

    this.sync = function (url, opts) {
        var autocreate = Opt(opts, 'autocreate', true);
        var creds = Opt(opts, 'creds', null);
        var target = new U1DBHTTPSyncTarget(url, creds);
        return new U1DBSynchronizer(this, target).sync(
            {autocreate: autocreate});
    }

}

/* Utils namespace */
U1DB.Utils = {};

(function() {

    U1DB.Utils.VarArgs = function(start, args) {
        var res = [];
        for (var i=start; i <  args.length; i++) {
            res.push(args[i]);
        }
        return res;
    }

    U1DB.Utils.Opt = function(opts, name, defl) {
        if (!(name in (opts||{}))) return defl;
        return opts[name];
    }

    /* simple deep equal, should be good enough for json compatible values */
    var deepEqual = function(a, b) {
        if (a === b) {
            return true;
        } else {
            if (typeof a != typeof b || typeof a != "object") {
                return false;
            }
            if (a === null || b === null) {
                return false;
            }
            if (a instanceof Array) {
                if (!(b instanceof Array)) {
                    return false;
                }
                if (a.length != b.length) {
                    return false;
                }
                for (var i=0; i < a.length; i++) {
                    if(!deepEqual(a[i], b[i])) {
                        return false;
                    }
                }
                return true;
            }
            var a_keys=[], b_keys=[];
            for (var a_key in a) {
                a_keys.push(a_key);
            }
            for (var b_key in b) {
                b_keys.push(b_key);
            }
            if (a_keys.length != b_keys.length) {
                return false;
            }
            for (var i=0; i <a_keys.length; i++) {
                var key = a_keys[i];
                if (!deepEqual(a[key], b[key])) {
                    return false;
                }
            }
            return true;
        }
    };

    U1DB.Utils.deepEqual = deepEqual;

})();


function U1DBHTTPSyncTarget(url, creds) {
    // xxx synchronous for now

    if (url && url.charAt(url.length-1) == '/') {
        url = url.substr(0, url.length-1);
    }

    var trace_hook_shallow = null;

    /* for testing */
    this._set_trace_hook_shallow = function(cb) {
        trace_hook_shallow = cb;
    }

    this.toString = function() {
        return "[U1DBHTTPSyncTarget " + url + "]";
    };

    var oauth_accessor = null;

    this.set_oauth_credentials = function(consumer_key, consumer_secret,
                                          token_key, token_secret) {
        oauth_accessor = {consumerKey: consumer_key,
                          consumerSecret: consumer_secret,
                          token: token_key,
                          tokenSecret: token_secret}
    }

    if (creds != null) {
        if (!('oauth' in creds)) {
            throw new U1DB.Errors.UnknownAuthMethod();
        }
        this.set_oauth_credentials(creds['oauth'].consumer_key,
                                   creds['oauth'].consumer_secret,
                                   creds['oauth'].token_key,
                                   creds['oauth'].token_secret)
    }

    var parseError = function(line) {
        var err = null;
        try {
            err = JSON.parse(line);
        } catch(e) {
        }
        if (err && typeof(err) == "object" && 'error' in err) {
            if (err.error in U1DB.wire_to_error) {
                throw new U1DB.wire_to_error[err.error](err.message);
            }
        }
    }

    var handleError = function(req) {
        parseError(req.responseText);
        throw new U1DB.Errors.HTTPError(req.status, req.responseText);
    }

    var prepareRequest = function(method, source_replica_uid, content_type) {
        var req = new XMLHttpRequest();
        var uri = url+'/sync-from/'+source_replica_uid;
        req.open(method, uri, false);
        if (content_type) {
            req.setRequestHeader("content-type", content_type);
        }
        if (oauth_accessor) {
            /* OAuth sign */
            var message = {
                method: method,
                action: uri
            };
            OAuth.setTimestampAndNonce(message);
            OAuth.completeRequest(message, oauth_accessor);
            var authorization_header = OAuth.getAuthorizationHeader(
                "" /* realm */, message.parameters)
            req.setRequestHeader('Authorization', authorization_header);
        }
        return req
    }

    this.get_sync_info = function(source_replica_uid) {
        var req = prepareRequest('GET', source_replica_uid);
        req.send();
        if (req.status == 200) {
            var res = JSON.parse(req.responseText);
            return [res['target_replica_uid'],
                    res['target_replica_generation'],
                    res['target_replica_transaction_id'],
                    res['source_replica_generation'],
                    res['source_transaction_id']]
        }
        handleError(req);
    }

    this.record_sync_info = function(source_replica_uid,
                                     source_replica_generation,
                                     source_transaction_id) {
        var req = prepareRequest('PUT', source_replica_uid,
                                 "application/json");
        if (trace_hook_shallow) {
            trace_hook_shallow('record_sync_info');
        }
        req.send(JSON.stringify({generation: source_replica_generation,
                                 transaction_id: source_transaction_id}));
        if (req.status == 200) {
            return;
        }
        handleError(req);
    }

    this._parse_sync_stream = function(data, return_doc_cb,
                                       ensure_callback) {
        var parts = data.split('\n');
        var nparts = parts.length;
        if (nparts && parts[nparts-1] == "") {
            nparts--;
        }
        if (nparts==0 || !/^\[\r?$/.test(parts[0])) {
            throw new U1DB.Errors.BrokenSyncStream();
        }
        var comma_line, res;
        if (nparts > 2) {
            var entry = null;
            function check_and_strip_comma(line) {
                var comma = false;
                var l = line.length;
                if (line.charAt(l-1) == '\r') {
                    l--;
                }
                if (line.charAt(l-1) == ',') {
                    l--;
                    comma = true;
                }
                return [comma, line.substring(0, l)];
            }
            comma_line = check_and_strip_comma(parts[1]);
            res = JSON.parse(comma_line[1]);
            if ('replica_uid' in res && ensure_callback != null) {
                ensure_callback(res['replica_uid']);
            }
            for (var j=2; j < nparts-1; j++) {
                if(!comma_line[0]) {
                    throw new U1DB.Errors.BrokenSyncStream();
                }
                comma_line = check_and_strip_comma(parts[j]);
                var entry = JSON.parse(comma_line[1]);
                var doc = new U1DBDocument(entry['id'], entry['rev']);
                doc.set_json(entry['content']);
                return_doc_cb(doc, entry['gen'], entry['trans_id']);
            }
        }
        if (!/^\]\r?$/.test(parts[nparts-1])) {
            parseError(parts[nparts-1]);
            throw new U1DB.Errors.BrokenSyncStream();
        }
        if(nparts <= 2|| comma_line[0]) {
            throw new U1DB.Errors.BrokenSyncStream();
        }
        return res;
    }

    this.sync_exchange = function(docs_by_generations, source_replica_uid,
                                  last_known_generation, last_known_trans_id,
                                  return_doc_cb, ensure_callback) {
        var req = prepareRequest('POST', source_replica_uid,
                                 "application/x-u1db-sync-stream");
        if (trace_hook_shallow) {
            trace_hook_shallow('sync_exchange');
        }
        var intro = {
            'last_known_generation': last_known_generation,
            'last_known_trans_id': last_known_trans_id
        }
        if (ensure_callback != null) {
            intro['ensure'] = true;
        }
        var entries = ['[\r\n' + JSON.stringify(intro)];
        docs_by_generations.forEach(function (doc_gen_trans_id) {
            entries.push(JSON.stringify({
                id: doc_gen_trans_id[0].doc_id,
                rev: doc_gen_trans_id[0].rev,
                content: doc_gen_trans_id[0].get_json(),
                gen: doc_gen_trans_id[1],
                trans_id: doc_gen_trans_id[2]
            }));
        });
        req.send(entries.join(',\r\n') + '\r\n]');
        if (req.status == 200) {
            var res = this._parse_sync_stream(req.responseText,
                                              return_doc_cb, ensure_callback);
            return [res['new_generation'], res['new_transaction_id']];
        }
        handleError(req);
    }

}

/* define errors */
(function() {
    U1DB.Errors = {};
    U1DB.wire_to_error = {};

    function U1DBError(message) {
        this.message = message || "";
        this.stack = (new Error()).stack;
    }
    U1DBError.prototype = new Error();
    U1DBError.prototype.name = "U1DBError";
    U1DB.Errors.U1DBError = U1DBError;

    function HTTPError(status, message) {
        U1DBError.apply(this, [message]);
        this.status = status;
    }
    HTTPError.prototype = new U1DBError();
    HTTPError.prototype.name = "HTTPError";
    U1DB.Errors.HTTPError = HTTPError;

    function defineErr(errorName, wire_description) {
        var myError = function(message) {
            U1DBError.apply(this, [message]);
        }
        myError.prototype = new U1DBError();
        myError.prototype.name = errorName;
        U1DB.Errors[errorName] = myError;
        if (wire_description != null) {
            myError.prototype.wire_description = wire_description;
            U1DB.wire_to_error[wire_description] = myError;
        }
    }

    defineErr('Unavailable', 'unavailable')
    defineErr('InvalidJSON', null)
    defineErr('DatabaseDoesNotExist', 'database does not exist')
    defineErr('UserQuotaExceeded', 'user quota exceeded')
    defineErr('IndexNameTakenError', null)
    defineErr('DocumentTooBig', 'document too big')
    defineErr('BrokenSyncStream', null)
    defineErr('InvalidGlobbing', null)
    defineErr('IndexDefinitionParseError', null)
    defineErr('InvalidDocId', 'invalid document id')
    defineErr('ConflictedDoc', null)
    defineErr('RevisionConflict', 'revision conflict')
    defineErr('InvalidValueForIndex', null)
    defineErr('InvalidTransactionId', 'invalid transaction id')
    defineErr('MissingDocIds', 'missing document ids')
    defineErr('InvalidGeneration', 'invalid generation')
    defineErr('Unauthorized', 'unauthorized')
    defineErr('DocumentDoesNotExist', 'document does not exist')
    defineErr('SubscriptionNeeded', 'user needs subscription')
    defineErr('IndexDoesNotExist', null)
    defineErr('DocumentAlreadyDeleted', 'document already deleted')
    defineErr('InvalidContent', null)
    defineErr('UnknownAuthMethod', null)

})()
